
#pragma once

#include <base/env.h>
#include <base/heap.h>

#include <tee_session/connection.h>
#include <tee/packet_allocator.h>
#include <google/protobuf/message.h>

namespace Tee {
	using Genode::Allocator_avl;
	using Genode::Env;
	using Genode::Heap;
	using Genode::error;
	using Genode::memset;

	using Packet_alloc_failed = Tee::Session::Tx::Source::Packet_alloc_failed;

	using google::protobuf::Message;

	class Proto_connection;
}


class Tee::Proto_connection
{
	private:
		Genode::Env&           _env;
		Genode::Heap           _heap { _env.ram(), _env.rm() };
		Genode::Allocator_avl  _tx_block_alloc { &_heap };
		Tee::Connection        _conn;
		char const*            _session_label;

	public:
		enum { DEFAULT_BUF_SIZE = Tee::Packet_allocator::DEFAULT_PACKET_SIZE * 128 };

		struct Serialize_failed    : Genode::Exception { };
		struct Deserialize_failed  : Genode::Exception { };
		struct Transmission_error  : Genode::Exception { };

		Proto_connection(Genode::Env&   env,
		                 char const*    label,
		                 Genode::size_t tx_buf_size = DEFAULT_BUF_SIZE,
		                 Genode::size_t rx_buf_size = DEFAULT_BUF_SIZE)
		: _env(env), _conn(env, &_tx_block_alloc, tx_buf_size, rx_buf_size, label),
			_session_label(label)

		{ }
		~Proto_connection() = default;

		/**
		 * Sends a protobuf message
		 *
		 * \param msg  The protobuf message to send.
		 * \param sid  The session id. Sinks need to reply using the received
		 *             session id so the source can track sessions.
		 *
		 * \throw Transmission_error  E.g. when there is no sink to send to.
		 * \throw Serialize_failed    Protobuf message could not be serialized.
		 * \throw Packet_alloc_failed Could not allocate packet on connection.
		 */
		void send(const google::protobuf::MessageLite& msg, const Tee::Session_id sid)
		{
			const size_t msg_size = msg.ByteSizeLong();
			auto tx_packet = _conn.tx()->alloc_packet(msg_size);
			tx_packet.length(msg_size);
			tx_packet.session_id(sid);
			auto tx_content = _conn.tx()->packet_content(tx_packet);
			Genode::memset(tx_content, 0, msg_size);

			if (msg.SerializeToArray(tx_content, msg_size)) {
				_conn.tx()->submit_packet(tx_packet);
				auto ack = _conn.tx()->get_acked_packet();
				if (ack.ok()) {
					_conn.tx()->release_packet(ack);
				} else {
					Genode::error("error in ack");
					throw Transmission_error();
				}
			} else {
				Genode::error("protobuf serialization failed");
				throw Serialize_failed();
			}
		}

		/**
		 * Blocking receive of a protobuf message
		 *
		 * \param msg  The protobuf message to receive
		 * \param sid  The session id. Sinks need to reply using the received
		 *             session id so the source can track sessions.
		 *
		 * \throw Deserialize_failed  Protobuf message could not be deserialized
		 */
		void receive(google::protobuf::MessageLite& msg, Tee::Session_id& sid)
		{
			auto rx_packet = _conn.rx()->get_packet();
			sid = rx_packet.session_id();

			auto rx_content = _conn.rx()->packet_content(rx_packet);
			_conn.rx()->acknowledge_packet(rx_packet);

			// deserialize protobuf
			if (!msg.ParseFromArray(rx_content, rx_packet.length())) {
				Genode::error("protobuf deserialization failed");
				throw Deserialize_failed();
			}
		}
};
